﻿using InTheHand.Net;
using InTheHand.Net.Bluetooth;
using InTheHand.Net.Sockets;
using System;
using System.Collections.Generic;
using System.Net.NetworkInformation;
using System.Text;

namespace ConsoleApp1
{
    class Program
    {
        static Dictionary<string, BluetoothDeviceInfo> _btList = new Dictionary<string, BluetoothDeviceInfo>();

        static void Main(string[] args)
        {
            // https://stackoverflow.com/questions/42701793/how-to-programatically-pair-a-bluetooth-device
            {
                BluetoothClient bluetoothClient;
                BluetoothComponent bluetoothComponent;

                bluetoothClient = new BluetoothClient();
                bluetoothComponent = new BluetoothComponent(bluetoothClient);

                bluetoothComponent.DiscoverDevicesProgress += bluetoothComponent_DiscoverDevicesProgress;
                bluetoothComponent.DiscoverDevicesComplete += bluetoothComponent_DiscoverDevicesComplete;

                bluetoothComponent.DiscoverDevicesAsync(255, false, true, true, false, null);
            }

            Console.WriteLine("Press any key to exit...");
            Console.ReadLine();
        }

        private static void bluetoothComponent_DiscoverDevicesComplete(object sender, DiscoverDevicesEventArgs e)
        {
            Console.WriteLine("_bluetoothComponent_DiscoverDevicesComplete");
        }

        private static void bluetoothComponent_DiscoverDevicesProgress(object sender, DiscoverDevicesEventArgs e)
        {
            foreach (var item in e.Devices)
            {
                if (_btList.ContainsKey(item.DeviceName) == true)
                {
                    continue;
                }

                _btList.Add(item.DeviceName, item);

                Console.WriteLine(item.DeviceName);
                Console.WriteLine($"\tConnected {item.Connected}");
                Console.WriteLine($"\tAuthenticated {item.Authenticated}");
                Console.WriteLine($"\tDeviceAddress {item.DeviceAddress}");

                foreach (var service in item.InstalledServices)
                {
                    Console.WriteLine("\t\t" + service);
                }

                if (string.Equals(item.DeviceName, "raspberrypi", StringComparison.OrdinalIgnoreCase) == true)
                {
                    ConnectTo(item);
                }
            }
        }

        private static void ConnectTo(BluetoothDeviceInfo item)
        {
            Guid serviceUUID = BluetoothService.TcpProtocol; // new Guid("00000004-0000-1000-8000-00805F9B34FB");

            BluetoothClient bluetoothClient = new BluetoothClient();
            bluetoothClient.Connect(item.DeviceAddress, serviceUUID);

            Console.WriteLine("Connected");
            bluetoothClient.Client.WriteString("Hello");
            string result = bluetoothClient.Client.ReadString();
            Console.WriteLine(result);

            bluetoothClient.Close();
            Console.WriteLine("Closed");
        }

        public static PhysicalAddress GetThisBTMacAddress()
        {
            foreach (NetworkInterface nic in NetworkInterface.GetAllNetworkInterfaces())
            {
                if (nic.Name == "Bluetooth Network Connection")
                {
                    return nic.GetPhysicalAddress();
                }
            }

            return null;
        }
    }
}


// 인터넷 연결이 설정된 Adapter profile
//var profile = InTheHand.Networking.Connectivity.NetworkInformation.GetInternetConnectionProfile();
//Console.WriteLine(profile);

// 이미 paired 상태의 device 열람
//string selector = InTheHand.Devices.Bluetooth.BluetoothDevice.GetDeviceSelector();
//foreach (var item in InTheHand.Devices.Enumeration.DeviceInformation.FindAll(selector))
//{
//    Console.WriteLine($"{item.Id}, {item.Name}, {item.Pairing}");
//}

// 이미 paired 상태의 device 열람
//InTheHand.Devices.Enumeration.DevicePicker picker = new InTheHand.Devices.Enumeration.DevicePicker();
//// picker.Filter
//var deviceInfo = picker.PickSingleDevice();
//Console.WriteLine(deviceInfo);
